@@ -1,7 +1,12 @@
 use crate::{iter::*, layout::*, util::*};
 
 use itertools::izip;
+#[cfg(feature = "parallel")]
 use rayon::prelude::*;
+#[cfg(target_arch = "x86")]
+use std::arch::x86::*;
+#[cfg(target_arch = "x86_64")]
+use std::arch::x86_64::*;
 
 pub fn apply_repulsion<T: Coord + std::fmt::Debug>(layout: &mut Layout<T>) {
 	let kr = layout.settings.kr.clone();
@@ -159,12 +164,8 @@ pub fn apply_repulsion_2d_parallel<T: Copy + Coord + std::fmt::Debug + Send + Sy
 	}
 }
 
+#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
 pub fn apply_repulsion_2d_simd_f64(layout: &mut Layout<f64>) {
-	#[cfg(target_arch = "x86")]
-	use std::arch::x86::*;
-	#[cfg(target_arch = "x86_64")]
-	use std::arch::x86_64::*;
-
 	for (n1, (n1_mass, n1_pos_s)) in layout.masses.iter().zip(layout.points.iter()).enumerate() {
 		let mut n2_iter = layout.points.iter();
 		let n1_mass = n1_mass + 1.;
@@ -244,8 +245,8 @@ pub fn apply_repulsion_2d_simd_f64(layout: &mut Layout<f64>) {
 					(layout.points.get(n2+1)[1] - layout.points.get(n1)[1]).powi(2),
 				));*/
 
-				let degs2 = (n1_mass * (layout.masses.get_unchecked(n2) + 1.)) as f64;
-				let degs3 = (n1_mass * (layout.masses.get_unchecked(n2 + 1) + 1.)) as f64;
+				let degs2 = n1_mass * (layout.masses.get_unchecked(n2) + 1.);
+				let degs3 = n1_mass * (layout.masses.get_unchecked(n2 + 1) + 1.);
 				let f = _mm_mul_pd(
 					_mm_div_pd(_mm_set_pd(degs3, degs2), d2),
 					_mm_set_pd(layout.settings.kr, layout.settings.kr),
@@ -321,12 +322,8 @@ pub fn apply_repulsion_2d_simd_f64(layout: &mut Layout<f64>) {
 	}
 }
 
+#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
 pub fn apply_repulsion_2d_simd_f32(layout: &mut Layout<f32>) {
-	#[cfg(target_arch = "x86")]
-	use std::arch::x86::*;
-	#[cfg(target_arch = "x86_64")]
-	use std::arch::x86_64::*;
-
 	for (n1, (n1_mass, n1_pos_s)) in layout.masses.iter().zip(layout.points.iter()).enumerate() {
 		let mut n2_iter = layout.points.iter();
 		let n1_mass = n1_mass + 1.0f32;
@@ -533,211 +530,209 @@ pub fn apply_repulsion_2d_simd_f32(layout: &mut Layout<f32>) {
 	}
 }
 
-/*pub fn apply_repulsion_2d_simd_f32_parallel(layout: &mut Layout<f32>) {
-	#[cfg(target_arch = "x86")]
-	use std::arch::x86::*;
-	#[cfg(target_arch = "x86_64")]
-	use std::arch::x86_64::*;
+#[cfg(all(feature = "parallel", any(target_arch = "x86", target_arch = "x86_64")))]
+pub fn apply_repulsion_2d_simd_f64_parallel(layout: &mut Layout<f64>) {
+	let chunk_size = layout.settings.chunk_size.unwrap();
+	let kr = layout.settings.kr;
+	for chunk_iter in layout.iter_par_simd_nodes::<2>(chunk_size) {
+		chunk_iter.for_each(|n1_iter| {
+			let n2_end = n1_iter.n2_end_ind;
 
-	let chunk_size = 128;
-	let n = layout.masses.len() - (layout.masses.len() % chunk_size);
+			for mut n1 in n1_iter {
+				let n1_mass = n1.mass + 1.;
+				let n1_pos = unsafe {
+					_mm256_set_pd(
+						*n1.pos.get_unchecked(1),
+						*n1.pos.get_unchecked(0),
+						*n1.pos.get_unchecked(1),
+						*n1.pos.get_unchecked(0),
+					)
+				};
 
-	for y0 in (0..n).step_by(chunk_size) {
-		(0..n - y0)
-			.into_par_iter()
-			.step_by(chunk_size)
-			.for_each(|x0| {
-				for (n1, n1_pos_s, n1_mass) in izip!(
-					x0..x0 + chunk_size,
-					layout.points.iter_from(x0),
-					layout.masses[x0..].iter(),
-				) {
-					//println!("n1: {}", n1);
+				// This loop iterates on nodes by 2
+				for n2 in &mut n1.n2_iter {
+					unsafe {
+						// [n2_x, n2_y, n3_x, n3_y]
+						let n23_pos = _mm256_loadu_pd(n2.pos);
 
-					let mut n2_iter = layout.points.iter_from(((n1 + 4) / 4 * 4).max(x0 + y0));
-					let n1_mass = n1_mass + 1.0f32;
-					let n1_speed = unsafe {
-						layout
-							.speeds
-							.points
-							.as_mut_ptr()
-							.add(n1 * layout.settings.dimensions)
-					};
-					let n1_pos = unsafe {
-						_mm256_set_ps(
-							n1_pos_s[1],
-							n1_pos_s[0],
-							n1_pos_s[1],
-							n1_pos_s[0],
-							n1_pos_s[1],
-							n1_pos_s[0],
-							n1_pos_s[1],
-							n1_pos_s[0],
-						)
-					};
+						// [dx(n1,n2), dx(n1,n2), dx(n1,n2), dx(n1,n3)]
+						let dxy = _mm256_sub_pd(n23_pos, n1_pos);
 
-					/*unsafe {
-						assert_eq!(std::mem::transmute::<__m256, (f32,f32,f32,f32,f32,f32,f32,f32)>(n1_pos), (
-							layout.points.get(n1)[0],
-							layout.points.get(n1)[1],
-							layout.points.get(n1)[0],
-							layout.points.get(n1)[1],
-							layout.points.get(n1)[0],
-							layout.points.get(n1)[1],
-							layout.points.get(n1)[0],
-							layout.points.get(n1)[1],
-						));
-					}*/
+						// ([dx(n1,n2)^2, dx(n1,n3)^2], [dy(n1,n2)^2, dy(n1,n3)^2])
+						let (dx2, dy2): (__m128d, __m128d) = std::mem::transmute(
+							_mm256_permute4x64_pd(_mm256_mul_pd(dxy, dxy), _MM_PERM_DBCA),
+						);
 
-					// TODO remaining (if n1+1 not multiple of 4)
+						// [d(n1,n2), d(n1,n3)]
+						let d2 = _mm_add_pd(dx2, dy2);
+						// TODO maybe check zero
 
-					for n2 in (((n1 + 4) / 4 * 4).max(x0 + y0)..x0 + y0 + chunk_size).step_by(4) {
-						//println!("n2: {} {} {} {} {}", n, x0, y0, n1, n2);
+						let degs2 = n1_mass * (*n2.mass + 1.);
+						let degs3 = n1_mass * (*n2.mass.add(1) + 1.);
+						let f = _mm_mul_pd(
+							_mm_div_pd(_mm_set_pd(degs3, degs2), d2),
+							_mm_set_pd(kr, kr),
+						);
 
-						unsafe {
-							// [n2_x, n2_y, n3_x, n3_y, n4_x, n4_y, n5_x, n5_y]
-							let n2345_pos = _mm256_loadu_ps(n2_iter.next_unchecked(4));
+						let (n1_speed_v, n23_speed_v): (__m128d, __m256d) =
+							(_mm_loadu_pd(n1.speed), _mm256_loadu_pd(n2.speed));
 
-							/*assert_eq!(std::mem::transmute::<__m256, (f32,f32,f32,f32,f32,f32,f32,f32)>(n2345_pos), (
-								layout.points.get(n2)[0],
-								layout.points.get(n2)[1],
-								layout.points.get(n2+1)[0],
-								layout.points.get(n2+1)[1],
-								layout.points.get(n2+2)[0],
-								layout.points.get(n2+2)[1],
-								layout.points.get(n2+3)[0],
-								layout.points.get(n2+3)[1],
-							));*/
+						let fxy = _mm256_mul_pd(
+							dxy,
+							_mm256_permute4x64_pd(_mm256_set_m128d(f, f), _MM_PERM_DBCA),
+						);
 
-							// [dx(n1,n2), dy(n1,n2), dx(n1,n3), dy(n1,n3), dx(n1,n4), dy(n1,n4), dx(n1,n5), dy(n1,n5)]
-							let dxy = _mm256_sub_ps(n2345_pos, n1_pos);
-
-							/*assert_eq!(std::mem::transmute::<__m256, (f32,f32,f32,f32,f32,f32,f32,f32)>(dxy), (
-								layout.points.get(n2)[0] - layout.points.get(n1)[0],
-								layout.points.get(n2)[1] - layout.points.get(n1)[1],
-								layout.points.get(n2+1)[0] - layout.points.get(n1)[0],
-								layout.points.get(n2+1)[1] - layout.points.get(n1)[1],
-								layout.points.get(n2+2)[0] - layout.points.get(n1)[0],
-								layout.points.get(n2+2)[1] - layout.points.get(n1)[1],
-								layout.points.get(n2+3)[0] - layout.points.get(n1)[0],
-								layout.points.get(n2+3)[1] - layout.points.get(n1)[1],
-							));*/
-
-							// ([dx(n1,n2)^2, dx(n1,n3)^2, dx(n1,n4)^2, dx(n1,n5)^2], [dy(n1,n2)^2, dy(n1,n3)^2, dy(n1,n4)^2, dy(n1,n5)^2])
-							let (dx2, dy2): (__m128, __m128) =
-								std::mem::transmute(_mm256_permutevar8x32_ps(
-									_mm256_mul_ps(dxy, dxy),
-									_mm256_set_epi32(7, 5, 3, 1, 6, 4, 2, 0),
-								));
-
-							/*assert_eq!((std::mem::transmute::<__m128, (f32,f32,f32,f32)>(dx2), std::mem::transmute::<__m128, (f32,f32,f32,f32)>(dy2)), (
-								(
-									(layout.points.get(n2)[0] - layout.points.get(n1)[0]).powi(2),
-									(layout.points.get(n2+1)[0] - layout.points.get(n1)[0]).powi(2),
-									(layout.points.get(n2+2)[0] - layout.points.get(n1)[0]).powi(2),
-									(layout.points.get(n2+3)[0] - layout.points.get(n1)[0]).powi(2)
+						_mm256_storeu_pd(n2.speed, _mm256_add_pd(n23_speed_v, fxy));
+						_mm_store_pd(
+							n1.speed,
+							_mm_sub_pd(
+								n1_speed_v,
+								_mm_add_pd(
+									_mm256_extractf128_pd(fxy, 1),
+									_mm256_extractf128_pd(fxy, 0),
 								),
-								(
-									(layout.points.get(n2)[1] - layout.points.get(n1)[1]).powi(2),
-									(layout.points.get(n2+1)[1] - layout.points.get(n1)[1]).powi(2),
-									(layout.points.get(n2+2)[1] - layout.points.get(n1)[1]).powi(2),
-									(layout.points.get(n2+3)[1] - layout.points.get(n1)[1]).powi(2)
-								)
-							));*/
-
-							// [d(n1,n2), d(n1,n3), d(n1,n4), d(n1,n5)]
-							let d2 = _mm_add_ps(dx2, dy2);
-							// TODO maybe check zero
-
-							/*assert_eq!(std::mem::transmute::<__m128, (f32,f32,f32,f32)>(d2), (
-								(layout.points.get(n2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2)[1] - layout.points.get(n1)[1]).powi(2),
-								(layout.points.get(n2+1)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+1)[1] - layout.points.get(n1)[1]).powi(2),
-								(layout.points.get(n2+2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+2)[1] - layout.points.get(n1)[1]).powi(2),
-								(layout.points.get(n2+3)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+3)[1] - layout.points.get(n1)[1]).powi(2),
-							));*/
-
-							let degs2 = n1_mass * (layout.masses.get_unchecked(n2) + 1.);
-							let degs3 = n1_mass * (layout.masses.get_unchecked(n2 + 1) + 1.);
-							let degs4 = n1_mass * (layout.masses.get_unchecked(n2 + 2) + 1.);
-							let degs5 = n1_mass * (layout.masses.get_unchecked(n2 + 3) + 1.);
-							let f = _mm_mul_ps(
-								_mm_div_ps(_mm_set_ps(degs5, degs4, degs3, degs2), d2),
-								_mm_set_ps(
-									layout.settings.kr,
-									layout.settings.kr,
-									layout.settings.kr,
-									layout.settings.kr,
-								),
-							);
-
-							/*assert_eq!(std::mem::transmute::<__m128, (f32,f32,f32,f32)>(f), (
-								degs2/((layout.points.get(n2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr,
-								degs3/((layout.points.get(n2+1)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+1)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr,
-								degs4/((layout.points.get(n2+2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+2)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr,
-								degs5/((layout.points.get(n2+3)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+3)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr,
-							));*/
-
-							let n2345_speed = layout
-								.speeds
-								.points
-								.as_mut_ptr()
-								.add(n2 * layout.settings.dimensions);
-							let (n1_speed_v, n2345_speed_v): (__m128, __m256) =
-								(_mm_loadu_ps(n1_speed), _mm256_loadu_ps(n2345_speed));
-
-							let fxy = _mm256_mul_ps(
-								dxy,
-								_mm256_permutevar8x32_ps(
-									_mm256_set_m128(f, f),
-									_mm256_set_epi32(7, 3, 6, 2, 5, 1, 4, 0),
-								),
-							);
-
-							/*assert_eq!(std::mem::transmute::<__m256, (f32,f32,f32,f32,f32,f32,f32,f32)>(fxy), (
-								degs2/((layout.points.get(n2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2)[0] - layout.points.get(n1)[0]),
-								degs2/((layout.points.get(n2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2)[1] - layout.points.get(n1)[1]),
-								degs3/((layout.points.get(n2+1)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+1)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2+1)[0] - layout.points.get(n1)[0]),
-								degs3/((layout.points.get(n2+1)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+1)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2+1)[1] - layout.points.get(n1)[1]),
-								degs4/((layout.points.get(n2+2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+2)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2+2)[0] - layout.points.get(n1)[0]),
-								degs4/((layout.points.get(n2+2)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+2)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2+2)[1] - layout.points.get(n1)[1]),
-								degs5/((layout.points.get(n2+3)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+3)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2+3)[0] - layout.points.get(n1)[0]),
-								degs5/((layout.points.get(n2+3)[0] - layout.points.get(n1)[0]).powi(2)+
-								(layout.points.get(n2+3)[1] - layout.points.get(n1)[1]).powi(2))*layout.settings.kr*(layout.points.get(n2+3)[1] - layout.points.get(n1)[1]),
-							));*/
-
-							_mm256_storeu_ps(n2345_speed, _mm256_add_ps(n2345_speed_v, fxy));
-							_mm_storeu_ps(
-								n1_speed,
-								_mm_sub_ps(
-									n1_speed_v,
-									_mm_add_ps(
-										_mm256_extractf128_ps(fxy, 1),
-										_mm256_extractf128_ps(fxy, 0),
-									),
-								),
-							);
-						}
+							),
+						);
 					}
 				}
-			});
+
+				// Remaining iteration
+				if n1.n2_iter.ind < n2_end {
+					let layout = unsafe { n1.n2_iter.layout.0.as_mut() };
+					let n2 = n1.n2_iter.ind;
+					let n2_pos = unsafe { layout.points.get_unchecked(n2) };
+
+					let dx = unsafe { *n2_pos.get_unchecked(0) - *n1.pos.get_unchecked(0) };
+					let dy = unsafe { *n2_pos.get_unchecked(1) - *n1.pos.get_unchecked(1) };
+
+					let d2 = dx * dx + dy * dy;
+					if d2.is_zero() {
+						continue;
+					}
+
+					let f = (n1_mass * (unsafe { layout.masses.get_unchecked(n2) } + 1.)) / d2 * kr;
+
+					let n2_speed = layout.speeds.get_mut(n2);
+					let vx = f * dx;
+					let vy = f * dy;
+					unsafe {
+						*n1.speed -= vx;
+						*n1.speed.add(1) -= vy;
+					}
+					unsafe { n2_speed.get_unchecked_mut(0) }.add_assign(vx);
+					unsafe { n2_speed.get_unchecked_mut(1) }.add_assign(vy);
+				}
+			}
+		});
 	}
-	// TODO remaining (if n not multiple of chunk_size)
-}*/
+}
+
+#[cfg(all(feature = "parallel", any(target_arch = "x86", target_arch = "x86_64")))]
+pub fn apply_repulsion_2d_simd_f32_parallel(layout: &mut Layout<f32>) {
+	let chunk_size = layout.settings.chunk_size.unwrap();
+	let kr = layout.settings.kr;
+	for chunk_iter in layout.iter_par_simd_nodes::<4>(chunk_size) {
+		chunk_iter.for_each(|n1_iter| {
+			let n2_end = n1_iter.n2_end_ind;
+
+			for mut n1 in n1_iter {
+				let n1_mass = n1.mass + 1.0f32;
+				let n1_pos = unsafe {
+					_mm256_set_ps(
+						*n1.pos.get_unchecked(1),
+						*n1.pos.get_unchecked(0),
+						*n1.pos.get_unchecked(1),
+						*n1.pos.get_unchecked(0),
+						*n1.pos.get_unchecked(1),
+						*n1.pos.get_unchecked(0),
+						*n1.pos.get_unchecked(1),
+						*n1.pos.get_unchecked(0),
+					)
+				};
+
+				// This loop iterates on nodes by 4
+				for n2 in &mut n1.n2_iter {
+					unsafe {
+						// [n2_x, n2_y, n3_x, n3_y, n4_x, n4_y, n5_x, n5_y]
+						let n2345_pos = _mm256_loadu_ps(n2.pos);
+
+						// [dx(n1,n2), dy(n1,n2), dx(n1,n3), dy(n1,n3), dx(n1,n4), dy(n1,n4), dx(n1,n5), dy(n1,n5)]
+						let dxy = _mm256_sub_ps(n2345_pos, n1_pos);
+
+						// ([dx(n1,n2)^2, dx(n1,n3)^2, dx(n1,n4)^2, dx(n1,n5)^2], [dy(n1,n2)^2, dy(n1,n3)^2, dy(n1,n4)^2, dy(n1,n5)^2])
+						let (dx2, dy2): (__m128, __m128) =
+							std::mem::transmute(_mm256_permutevar8x32_ps(
+								_mm256_mul_ps(dxy, dxy),
+								_mm256_set_epi32(7, 5, 3, 1, 6, 4, 2, 0),
+							));
+
+						// [d(n1,n2), d(n1,n3), d(n1,n4), d(n1,n5)]
+						let d2 = _mm_add_ps(dx2, dy2);
+						// TODO maybe check zero
+
+						let degs2 = n1_mass * (*n2.mass + 1.);
+						let degs3 = n1_mass * (*n2.mass.add(1) + 1.);
+						let degs4 = n1_mass * (*n2.mass.add(2) + 1.);
+						let degs5 = n1_mass * (*n2.mass.add(3) + 1.);
+						let f = _mm_mul_ps(
+							_mm_div_ps(_mm_set_ps(degs5, degs4, degs3, degs2), d2),
+							_mm_set_ps(kr, kr, kr, kr),
+						);
+
+						let (n1_speed_v, n2345_speed_v): (__m128, __m256) =
+							(_mm_loadu_ps(n1.speed), _mm256_loadu_ps(n2.speed));
+
+						let fxy = _mm256_mul_ps(
+							dxy,
+							_mm256_permutevar8x32_ps(
+								_mm256_set_m128(f, f),
+								_mm256_set_epi32(7, 3, 6, 2, 5, 1, 4, 0),
+							),
+						);
+
+						_mm256_storeu_ps(n2.speed, _mm256_add_ps(n2345_speed_v, fxy));
+						_mm_storeu_ps(
+							n1.speed,
+							_mm_sub_ps(
+								n1_speed_v,
+								_mm_add_ps(
+									_mm256_extractf128_ps(fxy, 1),
+									_mm256_extractf128_ps(fxy, 0),
+								),
+							),
+						);
+					}
+				}
+
+				// Remaining iterations
+				let layout = unsafe { n1.n2_iter.layout.0.as_mut() };
+				for n2 in n1.n2_iter.ind..n2_end {
+					let n2_pos = unsafe { layout.points.get_unchecked(n2) };
+
+					let dx = unsafe { *n2_pos.get_unchecked(0) - *n1.pos.get_unchecked(0) };
+					let dy = unsafe { *n2_pos.get_unchecked(1) - *n1.pos.get_unchecked(1) };
+
+					let d2 = dx * dx + dy * dy;
+					if d2.is_zero() {
+						continue;
+					}
+
+					let f = (n1_mass * (unsafe { layout.masses.get_unchecked(n2) } + 1.)) / d2 * kr;
+
+					let n2_speed = layout.speeds.get_mut(n2);
+					let vx = f * dx;
+					let vy = f * dy;
+					unsafe {
+						*n1.speed -= vx;
+						*n1.speed.add(1) -= vy;
+					}
+					unsafe { n2_speed.get_unchecked_mut(0) }.add_assign(vx);
+					unsafe { n2_speed.get_unchecked_mut(1) }.add_assign(vy);
+				}
+			}
+		});
+	}
+}
 
 pub fn apply_repulsion_3d<T: Copy + Coord + std::fmt::Debug>(layout: &mut Layout<T>) {
 	for (n1, (n1_mass, n1_pos)) in layout.masses.iter().zip(layout.points.iter()).enumerate() {
